from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, DTMLFile, ImageFile
from OFS.Folder import Folder
from Products.ZCatalog.CatalogPathAwareness import CatalogPathAware
from time import time
from DateTime import DateTime
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from workitem import workitem

def manage_addInstance(self,
                       process_id,
                       begin_activity_id,
                       customer='',
                       title='',
                       comments='',
                       priority=0,
                       REQUEST=None):
    i = instance(process_id, begin_activity_id, customer, title, comments, priority)
    if i.id in self.objectIds():
        instance_id = i.id
        ii=0
        while instance_id + str(ii) in self.objectIds():
            ii = ii + 1
        i.id = i.id + str(ii)
    self._setObject(i.id, i)
    if REQUEST: REQUEST.RESPONSE.redirect('manage_main')
    return i.id
    


class instance(CatalogPathAware, Folder):
    """ Even though it is called instance it is more than that: it is the collection
    of all instances related to a given process instance."""


    def __init__(self, process_id, activity_id, customer, title='', comments='', priority=0):
        self.customer = customer
        self.comments = comments
        self.id = customer + str(DateTime().timeTime())
        self.creation_time = DateTime()
        self.title = title
        self.priority = priority
        self.begin_process_id = process_id
        self.begin_activity_id = activity_id
        self.status = 'initiated' # initiated,running,active,complete,terminated,suspended
        self.old_status = '' #Used to remeber the status after a suspension
        #logging structure
        #each event has the form {'start':xxx, 'end':yyy, 'comment':zzz, 'actor':aaa}
        #'start' and 'end' are time in msec, 'actor' and 'comment' are string
        self.initiation_log = [] #1 log
        self.running_log = []
        self.activation_log = []
        self.completion_log = [] #1 log
        #statistic data in msec
        self.initiation_time = 0
        self.running_time = 0
        self.active_time = 0
        #log initialization
        self.initiation_log.append({'start':time(),'end':None,'comment':'creation','actor':''})


    security = ClassSecurityInfo()


    security.declareProtected('Use OpenFlow', 'History')
    index_html = PageTemplateFile('zpt/instance/History', globals())


    meta_type = 'Instance'


    manage_options = ({'label': 'History', 'action' : 'index_html'}, ) + Folder.manage_options[0:1] + Folder.manage_options[2:]


    def addWorkitem(self, process_id, activity_id, blocked, push_roles=[], pull_roles=[]):
        w_id = str(len(self.objectValues('Workitem')))
        w = workitem(w_id, self.id, process_id, activity_id, blocked,
                     push_roles=push_roles,
                     pull_roles=pull_roles)
        self._setObject(str(w.id), w)
        w.addEvent('creation')
        return w


    def getJoiningWorkitem(self, process_id, activity_id):
        w_list = filter (lambda x, pi=process_id, ai=activity_id : \
                         (x.status=='blocked') and \
                         (x.process_id==pi) and \
                         (x.activity_id==ai),
                         self.objectValues('Workitem'))
        if w_list:
            return w_list[0]
        else:
            return None

    def setStatus(self, status, comment='', actor=''):
        old_status = self.status
        new_status = status
        now = time()

        if old_status == 'initiated':
            self.initiation_log[-1]['end'] = now
            self.initiation_time += now - self.initiation_log[-1]['start']
            if new_status == 'running':
                self.running_log.append({'start':now,'end':None,'comment':comment,'actor':actor})

        if old_status == 'running':
            self.running_log[-1]['end'] = now
            self.running_time += now - self.running_log[-1]['start']
            if new_status == 'active':
                self.activation_log.append({'start':now,'end':None,'comment':comment,'actor':actor})
            if new_status == 'complete':
                self.completion_log.append({'start':now,'end':None,'comment':comment,'actor':actor})

        if old_status == 'active':
            self.activation_log[-1]['end'] = now
            self.active_time += now - self.activation_log[-1]['start']
            if new_status == 'running':
                self.running_log.append({'start':now,'end':None,'comment':comment,'actor':actor})

        self.status = status
        self.reindex_object()

    security.declareProtected('Use OpenFlow', 'getActiveWorkitems')
    def getActiveWorkitems(self):
        """ returns all active workitems"""
        return len(filter (lambda x: x.status == 'active', self.objectValues('Workitem')))


    def setPriority(self, value):
        self.priority = value
        self.reindex_object()

    def isActiveOrRunning(self):
        return self.status in ('active', 'running')

InitializeClass(instance)
