from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, DTMLFile
from DateTime import DateTime
from Products.ZCatalog.ZCatalog import ZCatalog
from instance import instance, manage_addInstance
from process import process, manage_addProcess
from OFS.Folder import Folder



from expression import exprNamespace
try:
    # do you have CMF?
    # if so use its Expression class
    from Products.CMFCore.Expression import Expression
except:
    # I guess you have no CMF...
    # here's a what you need:
    from expression import Expression

from Products.PageTemplates.PageTemplateFile import PageTemplateFile


def manage_addOpenflow(self, id, RESPONSE=None):
    """ adds a openflow folder in the zope repository """
    w = openflow(id)
    self._setObject(id, w)
    if RESPONSE is not None:
        RESPONSE.redirect('manage_main')


manage_addOpenflowForm = PageTemplateFile('zpt/openflow/manage_addOpenFlowForm', globals())

def manage_addProcessInstancesContainer(self, id):
    import Products
    if hasattr(Products, 'BTreeFolder2'):
        from Products.BTreeFolder2.BTreeFolder2 import BTreeFolder2
        ff = BTreeFolder2(id)
    elif hasattr(Products, 'BTreeFolder'):
        from Products.BTreeFolder.BTreeFolder import BTreeFolder
        ff = BTreeFolder()
        ff.id = id
    else:
        ff = Folder()
        ff.id = id
    ff.title = "Container for process instances"
    self._setObject(ff.id, ff)


class openflow(Folder):
    """ A openflow folder contains all the processes of the openflow """


    def __init__(self, id):
        self.id = id
        self._applications = {}
        self._activitiesPushableOnRole={}
        self._activitiesPullableOnRole={}
        # {'role':{'process':['activity',...]...}...}
        manage_addProcessInstancesContainer(self, 'ProcessInstances')
        catalog = ZCatalog('Catalog', 'Default OpenFlow Catalog')

        try:
            # zope 2.6.0 won't add these, so here they are
            catalog.addIndex('id', 'FieldIndex')
            catalog.addColumn('id')
            catalog.addIndex('meta_type', 'FieldIndex')
            catalog.addColumn('meta_type')
            catalog.addIndex('title', 'TextIndex')
            catalog.addColumn('title')
            catalog.addIndex('path', 'PathIndex')
            catalog.addColumn('path')
            v = Vocabulary(id='Vocabulary', title='')
            catalog._setObject(v.id, v)
        except:
            pass

        catalog.addIndex('description', 'FieldIndex')
        catalog.addColumn('description')

        catalog.addIndex('customer', 'FieldIndex')
        catalog.addColumn('customer')

        catalog.addIndex('creation_time', 'FieldIndex')
        catalog.addColumn('creation_time')

        catalog.addIndex('priority', 'FieldIndex')
        catalog.addColumn('priority')

        catalog.addIndex('status', 'FieldIndex')
        catalog.addColumn('status')

        catalog.addIndex('From', 'FieldIndex')
        catalog.addColumn('From')

        catalog.addIndex('To', 'FieldIndex')
        catalog.addColumn('To')

        catalog.addIndex('activity_id', 'FieldIndex')
        catalog.addColumn('activity_id')

        catalog.addIndex('process_id', 'FieldIndex')
        catalog.addColumn('process_id')

        catalog.addIndex('instance_id', 'FieldIndex')
        catalog.addColumn('instance_id')

        catalog.addIndex('workitems_from', 'KeywordIndex')
        catalog.addColumn('workitems_from')

        catalog.addIndex('workitems_to', 'KeywordIndex')
        catalog.addColumn('workitems_to')

        catalog.addIndex('actor', 'FieldIndex')
        catalog.addColumn('actor')

        catalog.addIndex('push_roles', 'KeywordIndex')
        catalog.addColumn('push_roles')

        catalog.addIndex('pull_roles', 'KeywordIndex')
        catalog.addColumn('pull_roles')

        self._setObject('Catalog', catalog)


    security = ClassSecurityInfo()


    security.declareObjectProtected('Use OpenFlow')

    
    security.declareProtected('Manage OpenFlow', 'manage_editActivitiesPushableOnRole')
    manage_editActivitiesPushableOnRole = PageTemplateFile('zpt/openflow/manage_editActivitiesPushableOnRole', globals())

    security.declareProtected('Manage OpenFlow', 'manage_editActivitiesPullableOnRole')
    manage_editActivitiesPullableOnRole = PageTemplateFile('zpt/openflow/manage_editActivitiesPullableOnRole', globals())

    security.declareProtected('Manage OpenFlow', 'manage_addApplicationForm')
    manage_addApplicationForm = PageTemplateFile('zpt/openflow/manage_addApplicationForm', globals())


    security.declareProtected('Manage OpenFlow', 'manage_editApplicationForm')
    manage_editApplicationForm = DTMLFile('dtml/openflow/manage_editApplicationForm', globals())


    security.declareProtected('Manage OpenFlow', 'manage_addInstanceForm')
    manage_addInstanceForm = PageTemplateFile('zpt/openflow/manage_addInstanceForm', globals())


    security.declareProtected('Manage OpenFlow', 'manage_addProcessForm')
    manage_addProcessForm = PageTemplateFile('zpt/openflow/manage_addProcessForm', globals())


    security.declareProtected('Manage OpenFlow', 'Roles')
    Roles = PageTemplateFile('zpt/openflow/Roles', globals())


    security.declareProtected('Manage OpenFlow', 'Applications')
    Applications = PageTemplateFile('zpt/openflow/Applications', globals())


    security.declareProtected('Use OpenFlow', 'index_html')
    index_html =  PageTemplateFile('zpt/openflow/WorkList', globals())


    security.declareProtected('Manage OpenFlow', 'Processes')
    Processes = PageTemplateFile('zpt/openflow/Processes', globals())


    security.declareProtected('Manage OpenFlow', 'Instances')
    Instances = PageTemplateFile('zpt/openflow/Instances', globals())


    security.declareProtected('Manage OpenFlow', 'Updates')
    Updates = PageTemplateFile('zpt/openflow/Updates', globals())
    

    security.declareProtected('Use OpenFlow', 'manage_pushWorkitem')
    manage_pushWorkitem = DTMLFile('dtml/openflow/manage_pushWorkitem', globals())


    security.declareProtected('Manage OpenFlow', 'chooseFallin')
    chooseFallin = DTMLFile('dtml/openflow/chooseFallin', globals())


    meta_type = 'OpenFlow'


    manage_options = ({'label': 'Worklist',
                       'action': 'index_html',
                       'help' : ('OpenFlow', 'openflow.stx')},
                      {'label': 'Roles',
                       'action': 'Roles',
                       'help' : ('OpenFlow', 'roles.stx')},
                      {'label': 'Applications',
                       'action': 'Applications',
                       'help' : ('OpenFlow', 'applications.stx')},
                      {'label': 'Process definitions',
                       'action': 'Processes',
                       'help' : ('OpenFlow', 'processes.stx')},
                      {'label':'Process instances',
                       'action':'Instances',
                       'help' : ('OpenFlow', 'processes.stx')},
                      {'label':'Updates',
                       'action':'Updates'}) + \
                       Folder.manage_options[0:1] + Folder.manage_options[2:]

    
    security.declarePrivate('manage_afterAdd')
    def manage_afterAdd(self, item, container):
        # this is to fix import and copy error:
        # Catalog is reset upon workflow creation
        # so that all cataloged items are recataloged upon their creation
        self.Catalog.manage_catalogClear()
        Folder.manage_afterAdd(self, item, container)
        

    security.declareProtected('Manage OpenFlow', 'editActivitiesPushableOnRole')
    def editActivitiesPushableOnRole(self, role, process, activities=[], REQUEST=[]):
        """ Edit the link between a role and activities of a process """
        if self._activitiesPushableOnRole.has_key(role) and \
               self._activitiesPushableOnRole[role].has_key(process):
            old_activities = self._activitiesPushableOnRole[role][process]
        else:
            old_activities = []
        removeList = [x for x in old_activities if x not in activities]
        if removeList:
            for i in self.Catalog.searchResults(meta_type='Workitem',process_id=process, activity_id=removeList):
                w = i.getObject() or getattr(self.getInstance(i.instance_id), i.id)
                if role in w.push_roles:
                    w.push_roles.remove(role)
                    w._p_changed = 1
                    w.reindex_object()
        addList = [x for x in activities if x not in old_activities]
        if addList:
            for i in self.Catalog.searchResults(meta_type='Workitem', process_id=process, activity_id=addList):
                w = i.getObject() or getattr(self.getInstance(i.instance_id), i.id)
                if role not in w.push_roles:
                    w.push_roles.append(role)
                    w._p_changed = 1
                    w.reindex_object()
        if activities:
            if not self._activitiesPushableOnRole.has_key(role):
                self._activitiesPushableOnRole[role] = {}
            self._activitiesPushableOnRole[role][process] = activities
        else:
            self.deleteProcessWithActivitiesPushableOnRole(role, process)
        self._p_changed = 1
        if REQUEST: REQUEST.RESPONSE.redirect('Roles')


    security.declareProtected('Manage OpenFlow', 'getActivitiesPushableOnRole')
    def getActivitiesPushableOnRole(self):
        """ """
        return self._activitiesPushableOnRole


    security.declareProtected('Manage OpenFlow', 'deleteProcessWithActivitiesPushableOnRole')
    def deleteProcessWithActivitiesPushableOnRole(self, role, process):
        """ Delete the link between a role and activities of a process """
        if self._activitiesPushableOnRole.has_key(role):
            if self._activitiesPushableOnRole[role].has_key(process):
                del self._activitiesPushableOnRole[role][process]
                if self._activitiesPushableOnRole[role] == {}:
                    self.deleteRoleWithActivitiesPushable(role)
        self._p_changed = 1


    security.declareProtected('Manage OpenFlow', 'deleteRoleWithActivitiesPushable')
    def deleteRoleWithActivitiesPushable(self, role):
        """ Delete a role """
        if self._activitiesPushableOnRole.has_key(role):
            del self._activitiesPushableOnRole[role]
        self._p_changed = 1


    security.declareProtected('Manage OpenFlow', 'addRoleWithActivitiesPushable')
    def addRoleWithActivitiesPushable(self, role, process):
        """ Add a role """
        self._activitiesPushableOnRole[role] = {}
        self._p_changed = 1


    security.declareProtected('Manage OpenFlow', 'editActivitiesPullableOnRole')
    def editActivitiesPullableOnRole(self, role, process, activities=[], REQUEST=[]):
        """ Edit the link between a role and activities of a process """
        if self._activitiesPullableOnRole.has_key(role) and \
               self._activitiesPullableOnRole[role].has_key(process):
            old_activities = self._activitiesPullableOnRole[role][process]
        else:
            old_activities = []
        removeList = [x for x in old_activities if x not in activities]
        if removeList:
            for i in self.Catalog.searchResults(meta_type='Workitem',process_id=process, activity_id=removeList):
                w = i.getObject() or getattr(self.getInstance(i.instance_id), i.id)
                if role in w.pull_roles:
                    w.pull_roles.remove(role)
                    w._p_changed = 1
                    w.reindex_object()
        addList = [x for x in activities if x not in old_activities]
        if addList:
            for i in self.Catalog.searchResults(meta_type='Workitem', process_id=process, activity_id=addList):
                w = i.getObject() or getattr(self.getInstance(i.instance_id), i.id)
                if role not in w.pull_roles:
                    w.pull_roles.append(role)
                    w._p_changed = 1
                    w.reindex_object()
        if activities:
            if not self._activitiesPullableOnRole.has_key(role):
                self._activitiesPullableOnRole[role] = {}
            self._activitiesPullableOnRole[role][process] = activities
        else:
            self.deleteProcessWithActivitiesPullableOnRole(role, process)
        self._p_changed = 1
        if REQUEST: REQUEST.RESPONSE.redirect('Roles')


    security.declareProtected('Manage OpenFlow', 'getActivitiesPullableOnRole')
    def getActivitiesPullableOnRole(self):
        """ """
        return self._activitiesPullableOnRole


    security.declareProtected('Manage OpenFlow', 'deleteProcessWithActivitiesPullableOnRole')
    def deleteProcessWithActivitiesPullableOnRole(self, role, process):
        """ Delete the link between a role and activities of a process """
        if self._activitiesPullableOnRole.has_key(role):
            if self._activitiesPullableOnRole[role].has_key(process):
                del self._activitiesPullableOnRole[role][process]
                if self._activitiesPullableOnRole[role] == {}:
                    self.deleteRoleWithActivitiesPullable(role)
        self._p_changed = 1


    security.declareProtected('Manage OpenFlow', 'deleteRoleWithActivitiesPullable')
    def deleteRoleWithActivitiesPullable(self, role):
        """ Delete a role """
        if self._activitiesPullableOnRole.has_key(role):
            del self._activitiesPullableOnRole[role]
        self._p_changed = 1


    security.declareProtected('Manage OpenFlow', 'addRoleWithActivitiesPullable')
    def addRoleWithActivitiesPullable(self, role, process):
        """ Add a role """
        self._activitiesPullableOnRole[role] = {}
        self._p_changed = 1


    security.declareProtected('Use OpenFlow', 'usersAssignableTo')
    def usersAssignableTo(self, process_id, activity_id):
        """ List all user name assignable to activity in the process """
        result=[]
        current=self
        apon = self._activitiesPullableOnRole
        pullable_roles = [r for r in apon.keys() \
                          if apon[r].has_key(process_id) and \
                          activity_id in apon[r][process_id]]
        while current is not None:
            if hasattr(current, 'acl_users'):
                for user in getattr(current, 'acl_users').getUsers():
                    name = user.getUserName()
                    roles_ok = [r for r in user.getRoles() if r in pullable_roles]
                    if roles_ok and name not in result:
                        result.append(name)
            try:
                current = current.aq_parent
            except:
                current = None
        return result


    security.declareProtected('Manage OpenFlow', 'addProcess')
    def addProcess(self,
                   id,
                   title="",
                   description="",
                   BeginEnd=None,
                   priority=0,
                   REQUEST=None):
        """ adds a new process """
        manage_addProcess(self,
                          id,
                          title,
                          description,
                          BeginEnd,
                          priority)
        if REQUEST: REQUEST.RESPONSE.redirect('Processes')


    security.declareProtected('Manage OpenFlow', 'deleteProcess')
    def deleteProcess(self, proc_ids=None, REQUEST=None):
        """ removes specified process """
        self.manage_delObjects(proc_ids)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'getPIContainer')
    def getPIContainer(self):
        """ returns the instance container """
        return getattr(self, "ProcessInstances", self)


    security.declareProtected('Use OpenFlow', 'getInstance')
    def getInstance(self, instance_id):
        """ returns a given instance """
        pi_container = self.getPIContainer()
        return getattr(pi_container, instance_id)


    security.declareProtected('Use OpenFlow', 'addInstance')
    def addInstance(self, process_id, customer='', comments='', title='', activation=0,
                    priority=0, REQUEST=None):
        """ adds an a process definition instance """
        pi_container = self.getPIContainer()
        begin_activity_id = getattr(self, process_id).begin
        instance_id = manage_addInstance(pi_container, process_id, begin_activity_id,
                                         customer, title, comments, priority)
        if activation:
            self.startInstance(instance_id)
        if REQUEST:
            REQUEST.RESPONSE.redirect('Instances')
        return instance_id


    security.declareProtected('Manage OpenFlow', 'deleteInstance')
    def deleteInstance(self, inst_ids=None, REQUEST=None):
        """ removes specified instance"""
        self.getPIContainer().manage_delObjects(inst_ids)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Manage OpenFlow', 'suspendInstance')
    def suspendInstance(self, instance_id=None, REQUEST=None):
        """ suspend a specified instance """
        instance = self.getInstance(instance_id)
        if instance.isActiveOrRunning() or instance.status == 'initiated':
            if REQUEST:
                actor=REQUEST.AUTHENTICATED_USER.getUserName()
            else:
                actor=''
            instance.old_status = instance.status
            instance.setStatus(status='suspended', actor=actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)
        

    security.declareProtected('Manage OpenFlow', 'suspendInstance')
    def resumeInstance(self, instance_id=None, REQUEST=None):
        """ suspend a specified instance """
        instance = self.getInstance(instance_id)
        if instance.status == 'suspended':
            if REQUEST:
                actor=REQUEST.AUTHENTICATED_USER.getUserName()
            else:
                actor=''
            instance.setStatus(status=instance.old_status, actor=actor)
            instance.old_status = ''
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)
        

    security.declareProtected('Manage OpenFlow', 'terminateInstance')
    def terminateInstance(self, instance_id=None, REQUEST=None):
        """ terminate a specified instance """
        instance = self.getInstance(instance_id)
        if instance.status != 'complete':
            if REQUEST:
                actor=REQUEST.AUTHENTICATED_USER.getUserName()
            else:
                actor=''
            instance.setStatus(status='terminated', actor=actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)
        

    security.declareProtected('Use OpenFlow', 'startInstance')
    def startInstance(self, instance_id, REQUEST=None):
        """ Starts the flowing of the process instance inside the process definition """
        instance = self.getInstance(instance_id)
        if REQUEST:
            actor=REQUEST.AUTHENTICATED_USER.getUserName()
        else:
            actor=''
        instance.setStatus(status='running', actor=actor)
        process_id = instance.begin_process_id
        activity_id = instance.begin_activity_id
        push_roles = self.getPushRoles(process_id, activity_id)
        pull_roles = self.getPullRoles(process_id, activity_id)
        w = instance.addWorkitem(process_id, activity_id, 0, push_roles, pull_roles)
        self.manageWorkitemCreation(instance_id, w.id)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    def linkWorkitems(self, instance_id, workitem_from_id, workitem_to_id_list):
        instance = self.getInstance(instance_id)
        workitem_from = getattr(instance, workitem_from_id)
        workitem_from.addTo(workitem_to_id_list)
        for w_id in workitem_to_id_list:
            w = getattr(instance, w_id)
            w.addFrom(workitem_from_id)
            w.setGraphLevel(workitem_from.graph_level + 1)


    def getPushRoles(self, process_id, activity_id):
        push_roles = []
        tmpRole = self._activitiesPushableOnRole
        for role in tmpRole.keys():
            if tmpRole[role].has_key(process_id):
                if activity_id in tmpRole[role][process_id]:
                    push_roles.append(role)
        return push_roles

    def getPullRoles(self, process_id, activity_id):
        pull_roles = []
        tmpRole = self._activitiesPullableOnRole
        for role in tmpRole.keys():
            if tmpRole[role].has_key(process_id):
                if activity_id in tmpRole[role][process_id]:
                    pull_roles.append(role)
        return pull_roles


    security.declareProtected('Manage OpenFlow', 'addApplication')
    def addApplication(self, name, link, REQUEST=None):
        """ adds an application declaration """
        if not name in self._applications.keys():
            self._applications[name] = {'url' : link}
            self._p_changed = 1
            if REQUEST:
                REQUEST.RESPONSE.redirect('Applications')


    security.declareProtected('Manage OpenFlow', 'deleteApplication')
    def deleteApplication(self, app_ids=None, REQUEST=None):
        """ removes an application """
        for name in app_ids:
            if name in self._applications.keys():
                del(self._applications[name])
        self._p_changed = 1
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Manage OpenFlow', 'editApplication')
    def editApplication(self, name, link, REQUEST=None):
        """ edits an application declaration """
        if name not in self._applications.keys():
            return
        self._applications[name] = {'url' : link}
        self._p_changed = 1
        if REQUEST:
            REQUEST.RESPONSE.redirect('Applications')


    security.declareProtected('Use OpenFlow', 'listApplications')
    def listApplications(self):
        """ List application declaration;
        returns a list of dictionaries with keys: name, link """
        return map(lambda x, self=self: {'name' : x,
                                         'link' : self._applications[x]['url']},
                                         self._applications.keys())


    security.declareProtected('Use OpenFlow', 'getApplicationUrl')
    def getApplicationUrl(self, instance_id, workitem_id):
        """ Return application definition URL relative to instance and workitem"""
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        application = activity.application
        if application in self._applications.keys():
            return self._applications[application]['url']
        else:
            return ""


    def getEnvironment(self, instance_id, workitem_id):
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, workitem_id)
        process_id = workitem.process_id
        activity_id = workitem.activity_id
        process = getattr(self, process_id)
        activity = getattr(process, activity_id)
        return instance, workitem, process, activity





    security.declareProtected('Use OpenFlow', 'assignWorkitem')
    def assignWorkitem(self, instance_id, workitem_id, actor, REQUEST=None):
        """ Assign the specified workitem of the specified instance to the specified actor (string)"""
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        user_is_ok = (REQUEST==None or
                      [r for r in REQUEST.AUTHENTICATED_USER.getRoles() if r in workitem.push_roles] or \
                      ([r for r in REQUEST.AUTHENTICATED_USER.getRoles() if r in workitem.pull_roles] and activity.isSelfAssignable()))
        workitem_is_ok = instance.isActiveOrRunning() and not workitem.status == 'completed' and workitem.actor == ''
        if user_is_ok and workitem_is_ok:
                workitem.assignTo(actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'unassignWorkitem')
    def unassignWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ Unassign the specified workitem """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, str(workitem_id))
        if instance.isActiveOrRunning() and workitem.status != 'completed':
            getattr(instance, workitem_id).assignTo('')
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'activateWorkitem')
    def activateWorkitem(self, instance_id, workitem_id, actor=None, REQUEST=None):
        """ declares the activation of the specified workitem of the given instance """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, str(workitem_id))
        if REQUEST: action_actor=REQUEST.AUTHENTICATED_USER.getUserName()
        else: action_actor=''
        if instance.isActiveOrRunning() and \
               workitem.status == 'inactive' and \
               not workitem.blocked:
            if actor is not None and workitem.actor == '':
                self.assignWorkitem(instance_id, workitem_id, actor)
            workitem.setStatus('active', actor=action_actor)
            instance.setStatus(status='active', actor=action_actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'inactivateWorkitem')
    def inactivateWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ declares the inactivation of the specified workitem of the given instance """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, str(workitem_id))
        if REQUEST: actor=REQUEST.AUTHENTICATED_USER.getUserName()
        else: actor=''
        if instance.isActiveOrRunning() and \
               workitem.status == 'active' and \
               not workitem.blocked:
            workitem.setStatus('inactive', actor=actor)
            if instance.getActiveWorkitems() == 0:
                instance.setStatus(status='running', actor=actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'suspendWorkitem')
    def suspendWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ declares the suspension of the specified workitem of the given instance """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, str(workitem_id))
        if REQUEST:
            actor = REQUEST.AUTHENTICATED_USER.getUserName()
        else:
            actor = ''
        if instance.isActiveOrRunning() and \
               workitem.status == 'inactive' and \
               not workitem.blocked:
            workitem.setStatus('suspended', actor=actor)
            if instance.getActiveWorkitems() == 0:
                instance.setStatus(status='running', actor=actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'resumeWorkitem')
    def resumeWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ declares the resumption of the specified workitem of the given instance """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, str(workitem_id))
        if REQUEST:
            actor = REQUEST.AUTHENTICATED_USER.getUserName()
        else:
            actor = ''
        if instance.isActiveOrRunning() and \
               workitem.status == 'suspended' and \
               not workitem.blocked:
            workitem.setStatus('inactive', actor=actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'completeWorkitem')
    def completeWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ declares the completion of the specified workitem of the given instance """
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        if REQUEST:
            actor = REQUEST.AUTHENTICATED_USER.getUserName()
        else:
            actor = ''
        if instance.isActiveOrRunning():
            workitem_return_id = None
            if workitem.status in ('active', 'fallout'):
                workitem.setStatus('complete', actor=actor)
                if instance.getActiveWorkitems() == 0:
                    instance.setStatus(status='running', actor=actor)
                if self.isEnd(workitem.process_id, workitem.activity_id):
                    subflow_workitem_id = self.getSubflowWorkitem(instance_id, workitem_id)
                    if subflow_workitem_id:
                        self.completeSubflow(instance_id, subflow_workitem_id)
                    else:
                        instance.setStatus(status='complete', actor=actor)
            if activity.isAutoFinish() and \
                   not process.end == activity.id and \
                   not activity.isSubflow():
                self.forwardWorkitem(instance_id, workitem_id)
            if REQUEST is not None: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    def isEnd(self, process_id, activity_id):
        process = getattr(self, process_id)
        return process.end == activity_id


    security.declareProtected('Use OpenFlow', 'getNextTransitions')
    def getNextTransitions(self, instance_id, workitem_id):
        """ Returns the list of transition that the given workitem (of the specified instance)
        will be routed on"""
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        transition_list = []
        split_mode = activity.split_mode
        activity_outgoing_transitions = filter(lambda x, activity_id=activity.id : x.From==activity_id, process.objectValues('Transition'))
        transition_condition_list = map(lambda x : {'transition_id' : x.id, 'condition' : x.condition}, activity_outgoing_transitions)
        if len(transition_condition_list) == 1:
            transition_list = [transition_condition_list[0]['transition_id']]
        else:
            if split_mode == 'and':
                transition_list = map(lambda x : x['transition_id'], transition_condition_list)
            elif split_mode == 'xor':
                for r in transition_condition_list:
                    expr=Expression(r['condition'])
		    ec=exprNamespace(instance=instance,
                                     workitem=workitem,
                                     process=process,
                                     activity=activity,
                                     openflow=self)
                    if expr(ec):
                        transition_list = [r['transition_id']]
                        break
        return transition_list


    def getDestinations(self, instance_id, workitem_id, path=None):
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        if path:
            transition_list = [path]
        else:
            transition_list = self.getNextTransitions(instance_id, workitem_id)
            if transition_list==[]: 
               raise "OpenflowNoRoute", "Workflow erroneous. The process don't know what to do now (no transitions to follow for this instance)."
        destinations = []
        for transition_id in transition_list:

            activity_to_id = getattr(process, transition_id).To
            activity_to = getattr(process, activity_to_id)
            if getattr(process, activity_to_id).join_mode=='and':
                blocked_init = activity_to.getIncomingTransitionsNumber() - 1
            else:
                blocked_init = 0

            destinations.append({'activity_to_id' : activity_to_id,
                                 'blocked_init' : blocked_init,
                                 'process_to_id' : process.id})

        return destinations


    security.declareProtected('Use OpenFlow', 'forwardWorkitem')
    def forwardWorkitem(self, instance_id, workitem_id, path=None, REQUEST=None):
        """ instructs openflow to forward the specified workitem """
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        destinations = self.getDestinations(instance_id, workitem_id, path)
        new_workitems = []
        if instance.isActiveOrRunning() and \
               (workitem.status == 'complete' and \
                (not workitem.workitems_to or activity.isSubflow())):
            activity_to_id_list = map(lambda x : x['activity_to_id'], destinations)
            workitem.addEvent('forwarded to '+ reduce(lambda x, y : x+', '+y, activity_to_id_list))
            workitem_to_id_list = []
            for d in destinations:
                w = instance.getJoiningWorkitem(d['process_to_id'], d['activity_to_id'])
                if w:
                    w.unblock()
                    workitem_to_id_list.append(w.id)
                else:
                    process_id = d['process_to_id']
                    activity_id = d['activity_to_id']
                    push_roles = self.getPushRoles(process_id, activity_id)
                    pull_roles = self.getPullRoles(process_id, activity_id)
                    w = instance.addWorkitem(process_id,
                                             activity_id,
                                             d['blocked_init'],
                                             push_roles,
                                             pull_roles)
                    workitem_to_id_list.append(w.id)
                if w.blocked == 0:
                    new_workitems.append(w.id)
                    w.addEvent('arrival from ' + workitem.activity_id)
            self.linkWorkitems(instance_id, workitem_id, workitem_to_id_list)
        for w in new_workitems:
            self.manageWorkitemCreation(instance_id, w)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'changeWorkitem')
    def changeWorkitem(self,
                       instance_id,
                       workitem_id,
                       push_roles = None,
                       pull_roles = None,
                       event_log = None,
                       process_id = None,
                       activity_id = None,
                       blocked = None,
                       priority = None,
                       workitems_from = None,
                       workitems_to = None,
                       status = None,
                       actor = None,
                       graph_level = None,
                       REQUEST = None):
        """ use this API to modify anything of a fallout workitem
            usable only if workitem status is 'fallout' 
        """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, workitem_id)
        if workitem.status=='fallout':
            if push_roles != None:
                workitem.push_roles = push_roles
            if pull_roles != None:
                workitem.pull_roles = pull_roles
            if event_log != None:
                workitem.event_log = event_log
            # the workitem.edit takes care of reindexing as well
            workitem.edit(process_id=process_id,
                          activity_id=activity_id,
                          blocked=blocked,
                          priority=priority,
                          workitems_from=workitems_from,
                          workitems_to=workitems_to,
                          status=status,
                          actor=actor,
                          graph_level=graph_level)
            workitem._p_changed = 1
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)
        
        
    security.declareProtected('Use OpenFlow', 'falloutWorkitem')
    def falloutWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ drops the workitem (of the specified instance) in exceptional handling """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, workitem_id)
        if REQUEST:
            actor = REQUEST.AUTHENTICATED_USER.getUserName()
        else:
            actor = ''
        workitem.setStatus('fallout', actor=actor)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Manage OpenFlow', 'fallinWorkitem')
    def fallinWorkitem(self, instance_id, workitem_id, process_id, activity_id, REQUEST=None, coming_from=None):
        """ the exceptional specified workitem (of the specified instance) will be put back in the activity
        specified by process_id and activity_id; workitem will still be in exceptional state:
        use endFallinWorkitem API to specify the end of the exceptional state"""
        instance = self.getInstance(instance_id)
        workitem_from = getattr(instance, workitem_id)
        push_roles = self.getPushRoles(process_id, activity_id)
        pull_roles = self.getPullRoles(process_id, activity_id)
        workitem_to = instance.addWorkitem(process_id, activity_id, 0, push_roles, pull_roles)
        self.linkWorkitems(instance_id, workitem_id, [workitem_to.id])
        event = 'fallin to activity ' + activity_id + ' in process ' + process_id + \
                ' (workitem ' + str(workitem_to.id) + ')'
        workitem_from.addEvent(event)
        event = 'fallin from activity ' + workitem_from.activity_id + \
                ' in process ' + workitem_from.process_id + \
                ' (workitem ' + str(workitem_id) + ')'
        workitem_to.addEvent(event)
        self.manageWorkitemCreation(instance_id, workitem_to.id)
        if REQUEST:
            if coming_from is not None:
                REQUEST.RESPONSE.redirect(coming_from)
            else:
                REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Manage OpenFlow', 'endFallinWorkitem')
    def endFallinWorkitem(self, instance_id, workitem_id, REQUEST=None):
        """ ends the exceptional state of the given workitem (of the specified instance) """
        instance = self.getInstance(instance_id)
        workitem = getattr(instance, workitem_id)
        workitem.addEvent('handled fallout')
        if not filter(lambda x: x['event'] == 'complete', workitem.getEventLog()):
            workitem.endFallin()
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Use OpenFlow', 'manageDummyActivity')
    def manageDummyActivity(self, instance_id, workitem_id):
        """  """
        self.activateWorkitem(instance_id, workitem_id, 'openflow_engine')
        self.completeWorkitem(instance_id, workitem_id)
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        if not self.isEnd(process.id, activity.id):
            self.forwardWorkitem(instance_id, workitem_id)


    security.declareProtected('Use OpenFlow', 'startAutomaticApplication')
    def startAutomaticApplication(self, instance_id, workitem_id):
        """  """
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        application_url = self.getApplicationUrl(instance_id,workitem_id)
        if application_url:
            args = {'instance_id':instance_id,
                    'workitem_id':workitem_id,
                    'openflow_id':self.id}
            apply(self.restrictedTraverse(application_url), (), args)
        self.activateWorkitem(instance_id, workitem_id, 'openflow_engine')
        self.completeWorkitem(instance_id, workitem_id)


    security.declarePrivate('callAutoPush')
    def callAutoPush(self, instance_id, workitem_id, REQUEST=None):
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        application_id = activity.push_application
        if application_id != "" and self._applications[application_id]['url']:
            application_url = self._applications[application_id]['url']
            func = self.restrictedTraverse(application_url)
            args = {'instance_id':instance.id,
                    'workitem_id':workitem.id,
                    'openflow_id':self.id}
            actor = apply(func,(), args)
            self.assignWorkitem(instance_id, workitem_id, actor)


    security.declarePrivate('startSubflow')
    def startSubflow(self, instance_id, workitem_id, REQUEST=None):
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        self.activateWorkitem(instance_id, workitem_id)
        self.assignWorkitem(instance_id, workitem_id, 'openflow_engine')
        subflow_id = activity.subflow
        begin_activity_id = getattr(self, subflow_id).begin
        push_roles = self.getPushRoles(subflow_id, begin_activity_id)
        pull_roles = self.getPullRoles(subflow_id, begin_activity_id)
        w = instance.addWorkitem(subflow_id, begin_activity_id, 0, push_roles, pull_roles)
        self.linkWorkitems(instance_id, workitem_id, [w.id])
        self.manageWorkitemCreation(instance_id, w.id)
        

    def getSubflowWorkitem(self, instance_id, workitem_id):
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        if activity.isSubflow():
            return workitem_id
        elif workitem.workitems_from == []:
            return None
        else:
            return self.getSubflowWorkitem(instance_id, workitem.workitems_from[0])


    def completeSubflow(self, instance_id, workitem_id):
        self.completeWorkitem(instance_id, workitem_id)
        self.forwardWorkitem(instance_id, workitem_id)


    def manageWorkitemCreation(self, instance_id, workitem_id):
        instance, workitem, process, activity = self.getEnvironment(instance_id, workitem_id)
        if instance.status in ('active', 'running'):
            if activity.isDummy():
                self.manageDummyActivity(instance_id, workitem_id)
            if activity.isAutoPush():
                self.callAutoPush(instance_id, workitem_id)
            if activity.isAutoStart():
                self.startAutomaticApplication(instance_id, workitem_id)
            if activity.isSubflow():
                self.startSubflow(instance_id, workitem_id)

    def runFixes(self, fixes=[], REQUEST=None):
        """ runs the selected fixes """
        for fix in fixes:
            if fix[0:3]=='fix' and hasattr(self, fix) and \
                   hasattr(getattr(self, fix), '__doc__'):
                getattr(self, fix)()
        if REQUEST != None:
            REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)

    def getNeededFixes(self):
        """ finds the fixes the current openflow needs """
        result = []
        if not hasattr(self, 'ProcessInstances'):
            result.append('fixCreateProcessInstancesContainer')
        if self.objectIds('Instance') and hasattr(self, 'ProcessInstances'):
            result.append('fixMoveInstancesInTheirContainer')
        return result

    def fixCreateProcessInstancesContainer(self):
        """ this is a fix moving OF 1.0.6 to OF 1.1 """
        manage_addProcessInstancesContainer(self, 'ProcessInstances')

    def fixMoveInstancesInTheirContainer(self):
        """ this is a fix moving OF 1.0.6 to OF 1.1 """
        from Acquisition import aq_base, aq_inner, aq_parent
        pic = self.getPIContainer()
        cb = self.manage_cutObjects(self.objectIds('Instance'))
        pic.manage_pasteObjects(cb)


InitializeClass(openflow)
