from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, DTMLFile, ImageFile
from OFS.Folder import Folder, manage_addFolder
from DateTime import DateTime
from Products.ZCatalog.CatalogPathAwareness import CatalogPathAware
from activity import activity
from transition import transition
from Products.PageTemplates.PageTemplateFile import PageTemplateFile

from Globals import package_home

def manage_addProcess(self, id,
                      title,
                      description,
                      BeginEnd,
                      priority,
                      REQUEST=None):
    p = process(id,
                title,
                description,
                BeginEnd,
                priority)
    self._setObject(id, p)
    if REQUEST: REQUEST.RESPONSE.redirect('manage_main')
    

class process(CatalogPathAware, Folder):
    """ A process is a collection of activities and transitions.
    The process map is given by the linking of activities by transitions.
    Each process instance is described by a instance"""

    def __init__(self,
                 id,
                 title="",
                 description="",
                 BeginEnd=None,
                 priority=0):
        self.id = id
        self.title = title
        self.description = description
        self.created = DateTime()
        self.priority = priority
        if BeginEnd:
            self.addActivity('Begin')
            self.addActivity('End')
            self.begin = 'Begin'
            self.end = 'End'
        else:
            self.begin = ''
            self.end = ''


    security = ClassSecurityInfo()


    security.declareProtected('Manage OpenFlow', 'manage_addActivityForm')
    manage_addActivityForm = DTMLFile('dtml/process/manage_addActivityForm', globals())


    security.declareProtected('Manage OpenFlow', 'manage_addTransitionForm')
    manage_addTransitionForm = DTMLFile('dtml/process/manage_addTransitionForm', globals())


    security.declareProtected('Manage OpenFlow', 'Map')
    index_html = DTMLFile('dtml/process/Map', globals())


    security.declareProtected('Manage OpenFlow', 'Setting')
    Setting = PageTemplateFile('zpt/process/Setting', globals())


    meta_type = 'Process'


    manage_options = ({'label' : 'Map', 'action' : 'index_html'},
                      {'label' : 'Setting', 'action' : 'Setting'}
                      ) + Folder.manage_options[0:1] + Folder.manage_options[2:]


    security.declareProtected('Manage OpenFlow', 'edit')
    def edit(self,
             begin=None,
             end=None,
             title=None,
             description=None, 
             priority=None,
             REQUEST=None):
        """ changes the process settings """
        if title:
            self.title = title
        if description:
            self.description = description
        if begin:
            self.begin = begin
        if end:
            self.end = end
        if priority:
            self.priority = priority
        if REQUEST:
            return self.Setting(self,REQUEST,manage_tabs_message="Changed")


    security.declareProtected('Manage OpenFlow', 'addActivity')
    def addActivity(self,
                    id,
                    split_mode='and',
                    join_mode='and',
                    self_assignable=1,
                    start_mode=0,
                    finish_mode=0,
                    subflow='',
                    push_application='',
                    application='',
                    title='',
                    parameters='',
                    description='',
                    kind = 'standard',
                    REQUEST=None):
        """ adds the activity and eventually sets the process begin and end activity """
        a = activity(id=id,
                     join_mode=join_mode,
                     split_mode=split_mode,
                     self_assignable=self_assignable,
                     start_mode=start_mode,
                     finish_mode=finish_mode,
                     subflow=subflow,
                     push_application=push_application,
                     application=application,
                     title=title,
                     parameters=parameters,
                     description=description,
                     kind=kind)
        self._setObject(id, a)
        if REQUEST: REQUEST.RESPONSE.redirect('index_html')


    security.declareProtected('Manage OpenFlow', 'addTransition')
    def addTransition(self, id, From, To, condition=None, REQUEST=None):
        """ adds a transition """
        t = transition(id, From, To, condition)
        self._setObject(t.id, t)
        if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


    security.declareProtected('Manage OpenFlow', 'manage_delObjects')
    def manage_delObjects(self, ids=[], REQUEST=None):
      """ override default method to handle better the redirection """
      Folder.manage_delObjects(self, ids)
      if REQUEST: REQUEST.RESPONSE.redirect(REQUEST.HTTP_REFERER)


InitializeClass(process)
